{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from matplotlib import pyplot as plt\n",
    "from PIL import Image\n",
    "import numpy as np\n",
    "from skimage.measure import block_reduce\n",
    "import requests\n",
    "\n",
    "\n",
    "with open('s34_e150_1arc_v3.tif','wb') as f:\n",
    "    f.write(requests.get(\n",
    "        'http://sendimage.whu.edu.cn/res/DEM_share/SRTM1/S60,E150/s34_e150_1arc_v3.tif'\n",
    "    ).content)\n",
    "    f.close()\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "im = np.asarray(Image.open('s34_e150_1arc_v3.tif'))\n",
    "plt.imshow(im)\n",
    "plt.show()\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "\"\"\"UnionFind.py\n",
    "\n",
    "Union-find data structure. Based on Josiah Carlson's code,\n",
    "http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/215912\n",
    "with significant additional changes by D. Eppstein.\n",
    "\"\"\"\n",
    "\n",
    "\n",
    "class UnionFind:\n",
    "\n",
    "    \"\"\"Union-find data structure.\n",
    "\n",
    "    Each unionFind instance X maintains a family of disjoint sets of\n",
    "    hashable objects, supporting the following two methods:\n",
    "\n",
    "    - X[item] returns a name for the set containing the given item.\n",
    "      Each set is named by an arbitrarily-chosen one of its members; as\n",
    "      long as the set remains unchanged it will keep the same name. If\n",
    "      the item is not yet part of a set in X, a new singleton set is\n",
    "      created for it.\n",
    "\n",
    "    - X.union(item1, item2, ...) merges the sets containing each item\n",
    "      into a single larger set.  If any item is not yet part of a set\n",
    "      in X, it is added to X as one of the members of the merged set.\n",
    "    \"\"\"\n",
    "\n",
    "    def __init__(self):\n",
    "        \"\"\"Create a new empty union-find structure.\"\"\"\n",
    "        self.weights = {}\n",
    "        self.parents = {}\n",
    "\n",
    "    def add(self, object, weight):\n",
    "        if object not in self.parents:\n",
    "            self.parents[object] = object\n",
    "            self.weights[object] = weight\n",
    "\n",
    "    def __contains__(self, object):\n",
    "        return object in self.parents\n",
    "\n",
    "    def __getitem__(self, object):\n",
    "        \"\"\"Find and return the name of the set containing the object.\"\"\"\n",
    "\n",
    "        # check for previously unknown object\n",
    "        if object not in self.parents:\n",
    "            assert(False)\n",
    "            self.parents[object] = object\n",
    "            self.weights[object] = 1\n",
    "            return object\n",
    "\n",
    "        # find path of objects leading to the root\n",
    "        path = [object]\n",
    "        root = self.parents[object]\n",
    "        while root != path[-1]:\n",
    "            path.append(root)\n",
    "            root = self.parents[root]\n",
    "\n",
    "        # compress the path and return\n",
    "        for ancestor in path:\n",
    "            self.parents[ancestor] = root\n",
    "        return root\n",
    "\n",
    "    def __iter__(self):\n",
    "        \"\"\"Iterate through all items ever found or unioned by this structure.\n",
    "\n",
    "        \"\"\"\n",
    "        return iter(self.parents)\n",
    "\n",
    "    def union(self, *objects):\n",
    "        \"\"\"Find the sets containing the objects and merge them all.\"\"\"\n",
    "        roots = [self[x] for x in objects]\n",
    "        heaviest = max([(self.weights[r], r) for r in roots])[1]\n",
    "        for r in roots:\n",
    "            if r != heaviest:\n",
    "                self.parents[r] = heaviest\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from dataclasses import dataclass\n",
    "from tqdm import tqdm\n",
    "\n",
    "\n",
    "@dataclass\n",
    "class Peak:\n",
    "    birth_point: tuple\n",
    "    birth_level: float\n",
    "    persistence: float\n",
    "    death_point: tuple\n",
    "\n",
    "\n",
    "class Mountaineer:\n",
    "    \"\"\"\n",
    "    Finds the peaks in the image.\n",
    "    \n",
    "    Uses concepts from persistent homology, and the UnionFind data structure.\n",
    "    Adapted from:  https://git.sthu.org/?p=persistence.git;a=summary\n",
    "    \"\"\"\n",
    "    def __init__(self, image):\n",
    "        self.image = image\n",
    "        self.h, self.w = image.shape\n",
    "\n",
    "        self.uf = UnionFind()\n",
    "\n",
    "    def get_comp_birth(self, p):\n",
    "        pp = self.uf[p]\n",
    "        y, x = pp[0], pp[1]\n",
    "\n",
    "        return self.image[y, x]\n",
    "\n",
    "    def survey_peaks(self):\n",
    "        # Get sorted array of indices, highest first\n",
    "        idxs_sorted = np.dstack(\n",
    "            np.unravel_index(np.argsort(self.image.ravel()), self.image.shape)\n",
    "        )[0, ...][::-1, :]\n",
    "        \n",
    "        idx_im = np.indices(self.image.shape)\n",
    "\n",
    "        groups0 = {}\n",
    "        \n",
    "        for i, p in tqdm(enumerate(idxs_sorted)):\n",
    "            p = (p[0], p[1])\n",
    "            y, x = p[0], p[1]\n",
    "            val = self.image[y, x]\n",
    "            if val == -np.inf: continue\n",
    "            \n",
    "            _y, y_ = max(y - 1, 0), min(y + 1, self.h)\n",
    "            _x, x_ = max(x - 1, 0), min(x + 1, self.w)\n",
    "            \n",
    "            neighbours = idx_im[:, _y:y_+1, _x:x_+1].transpose((2, 1, 0)).reshape((-1, 2))\n",
    "            neighbours = list(map(tuple, neighbours))\n",
    "\n",
    "            ni = [self.uf[q] for q in neighbours if q in self.uf]\n",
    "            nc = sorted([(self.get_comp_birth(q), q) for q in set(ni)], reverse=True)\n",
    "\n",
    "            if i == 0:\n",
    "                groups0[p] = Peak(birth_point=None, birth_level=val, persistence=val, death_point=None)\n",
    "\n",
    "            self.uf.add(p, -i)\n",
    "\n",
    "            if len(nc) > 0:\n",
    "                oldp = nc[0][1]\n",
    "                self.uf.union(oldp, p)\n",
    "                # Merge all others with oldp\n",
    "                for bl, q in nc[1:]:\n",
    "                    if self.uf[q] not in groups0:\n",
    "                        groups0[self.uf[q]] = Peak(birth_point=None, birth_level=bl, persistence=bl-val, death_point=p)\n",
    "\n",
    "                    self.uf.union(oldp, q)\n",
    "\n",
    "        def add_bp(peak: Peak, bp):\n",
    "            peak.birth_point = bp\n",
    "            return peak\n",
    "        \n",
    "        g = groups0\n",
    "        groups = [add_bp(v, k) for k, v in groups0.items()]\n",
    "        groups.sort(key=lambda g: g.persistence, reverse=True)\n",
    "\n",
    "        return groups\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "peaks = Mountaineer(block_reduce(im, (4, 4), np.mean)).survey_peaks()\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "im = block_reduce(im, (4, 4), np.mean)\n",
    "\n",
    "fig = plt.figure(figsize=(5, 10), dpi=120)\n",
    "plt.imshow(im)\n",
    "\n",
    "ax = fig.add_subplot(111)\n",
    "for i, peak in enumerate(peaks):\n",
    "    p_birth, bl, pers, p_death = peak.birth_point, peak.birth_level, peak.persistence, peak.death_point\n",
    "    if pers <= 20.0:\n",
    "        continue\n",
    "    y, x = p_birth\n",
    "    ax.plot([x], [y], '.', c='b')\n",
    "    ax.text(x, y+0.25, str(i+1), color='red')\n",
    "    if i >= 20: break\n",
    "    \n",
    "ax.set_xlim((0,im.shape[1]))\n",
    "ax.set_ylim((0,im.shape[0]))\n",
    "plt.gca().invert_yaxis()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import matplotlib.pyplot as plt\n",
    "import numpy as np\n",
    "from matplotlib import cm\n",
    "\n",
    "img = im\n",
    "\n",
    "\n",
    "def get(p):\n",
    "    if p == None:\n",
    "        return img.min()\n",
    "    return img[p[0], p[1]]\n",
    "\n",
    "pts = [p for p in peaks]\n",
    "pts.sort(key=lambda x: get(x.birth_point), reverse=True)\n",
    "\n",
    "persistences = np.asarray([p.persistence for p in pts])\n",
    "_max, _min = persistences.max(), persistences.min()\n",
    "\n",
    "pts = list(filter(lambda x: x.persistence > 150, pts, ))\n",
    "\n",
    "persistences = np.asarray([p.persistence for p in pts])\n",
    "colors = cm.viridis((persistences - _min) / (_max - _min))[::-1]\n",
    "\n",
    "\n",
    "begin = np.asarray([get(p.birth_point) for p in pts])[::-1]\n",
    "end = np.asarray([get(p.death_point) for p in pts])[::-1]\n",
    "p_sorted = [p for p in peaks]\n",
    "p_sorted.sort(key=lambda x: x.persistence, reverse=True)\n",
    "tags = {p.birth_point: idx for idx, p in enumerate(p_sorted)}\n",
    "tags = [str(tags[p.birth_point]+1) for p in pts][::-1]\n",
    "\n",
    "fig = plt.figure(figsize=(10, 15))\n",
    "plot = plt.scatter(begin, begin, c = begin, cmap = 'viridis')\n",
    "plt.clf()\n",
    "plt.grid(linestyle=':')\n",
    "\n",
    "cbar = plt.colorbar(plot, aspect=70)\n",
    "\n",
    "plot = plt.barh(range(len(begin)),  end-begin, left=begin, color=colors)\n",
    "plt.gca().invert_xaxis()\n",
    "\n",
    "plt.margins(y=0.01)\n",
    "plt.xlabel('Elevation (m)', fontsize=16)\n",
    "plt.ylabel('Birth and Death of Individual Peaks\\n(label refers to persistence rank)', fontsize=16)\n",
    "cbar.ax.set_ylabel(\"Persistence (m)\", fontsize=16)\n",
    "plt.title(\"Bar Chart of 66-most Persistent Peaks\", fontsize=16)\n",
    "plt.yticks(range(len(begin)), tags)\n",
    "plt.tight_layout()\n",
    "plt.savefig('barchart.png', dpi=120)\n",
    "plt.show()\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.5"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
